from arcana.study.base import Study, StudyMetaClass
from arcana.data import FilesetSpec, FieldSpec, FilesetInputSpec
from banana.file_format import (
    nifti_gz_format, text_format, text_matrix_format, directory_format,
    list_mode_format)
from banana.interfaces.sklearn import FastICA
from banana.interfaces.ants import AntsRegSyn
import os
from banana.requirement import fsl_req, mrtrix_req
from banana.interfaces.custom.pet import PreparePetDir
from banana.interfaces.custom.dicom import PetTimeInfo
from arcana.study import ParamSpec
from banana.interfaces.custom.pet import (
    PrepareUnlistingInputs, PETListModeUnlisting, SSRB, MergeUnlistingOutputs)
from banana.requirement import stir_req


template_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__).split('arcana')[0],
                 'arcana', 'reference'))


class PetStudy(Study, metaclass=StudyMetaClass):

    add_param_specs = [ParamSpec('ica_n_components', 2),
                        ParamSpec('ica_type', 'spatial'),
                        ParamSpec('norm_transformation', 's'),
                        ParamSpec('norm_dim', 3),
                        ParamSpec('norm_template',
                                      os.path.join(template_path,
                                                   'PET_template.nii.gz')),
                        ParamSpec('crop_xmin', 100),
                        ParamSpec('crop_xsize', 130),
                        ParamSpec('crop_ymin', 100),
                        ParamSpec('crop_ysize', 130),
                        ParamSpec('crop_zmin', 20),
                        ParamSpec('crop_zsize', 100),
                        ParamSpec('image_orientation_check', False)]

    add_data_specs = [
        FilesetInputSpec('list_mode', list_mode_format),
        FilesetSpec('registered_volumes', nifti_gz_format, optional=True),
        FilesetSpec('pet_image', nifti_gz_format, optional=True),
        FilesetSpec('pet_data_dir', directory_format),
        FilesetSpec('pet_recon_dir', directory_format),
        FilesetSpec('pet_recon_dir_prepared', directory_format,
                    'pet_data_preparation_pipeline'),
        FilesetSpec('decomposed_file', nifti_gz_format, 'ICA_pipeline'),
        FilesetSpec('timeseries', nifti_gz_format, 'ICA_pipeline'),
        FilesetSpec('mixing_mat', text_format, 'ICA_pipeline'),
        FilesetSpec('registered_volume', nifti_gz_format,
                    'Image_normalization_pipeline'),
        FilesetSpec('warp_file', nifti_gz_format,
                    'Image_normalization_pipeline'),
        FilesetSpec('invwarp_file', nifti_gz_format,
                    'Image_normalization_pipeline'),
        FilesetSpec('affine_mat', text_matrix_format,
                    'Image_normalization_pipeline'),
        FieldSpec('pet_duration', dtype=int,
                  'pet_time_info_extraction_pipeline'),
        FieldSpec('pet_end_time', dtype=str,
                  'pet_time_info_extraction_pipeline'),
        FieldSpec('pet_start_time', dtype=str,
                  'pet_time_info_extraction_pipeline'),
        FieldSpec('time_offset', int),
        FieldSpec('temporal_length', float),
        FieldSpec('num_frames', int),
        FilesetSpec('ssrb_sinograms', directory_format,
                    'sinogram_unlisting_pipeline')]

    def ICA_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='ICA',
            desc=('Decompose a 4D fileset into a set of independent '
                  'components using FastICA'),
            citations=[],
            **kwargs)

        pipeline.add(
            'ICA',
            FastICA(
                n_components=self.parameter('ica_n_components'),
                ica_type=self.parameter('ica_type')),
            inputs={
                'volume': ('registered_volumes', nifti_gz_format)},
            ouputs={
                'decomposed_file': ('ica_decomposition', nifti_gz_format),
                'timeseries': ('ica_timeseries', nifti_gz_format),
                'mixing_mat': ('mixing_mat', text_format)})

        return pipeline

    def Image_normalization_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='Image_registration',
            desc=('Image registration to a template using ANTs'),
            citations=[],
            **kwargs)

        pipeline.add(
            'ANTs',
            AntsRegSyn(
                out_prefix='vol2template',
                num_dimensions=self.parameter('norm_dim'),
                num_threads=self.processor.num_processes,
                transformation=self.parameter('norm_transformation'),
                ref_file=self.parameter('norm_template')),
            inputs={
                'input_file': ('pet_image', nifti_gz_format)},
            ouputs={
                'registered_volume': ('reg_file', nifti_gz_format),
                'warp_file': ('warp_file', nifti_gz_format),
                'invwarp_file': ('inv_warp', nifti_gz_format),
                'affine_mat': ('regmat', text_matrix_format)})

        return pipeline

    def pet_data_preparation_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='pet_data_preparation',
            desc=("Given a folder with reconstructed PET data, this "
                  "pipeline will prepare the data for the motion "
                  "correction"),
            citations=[],
            **kwargs)

        pipeline.add(
            'prepare_pet',
            PreparePetDir(
                image_orientation_check=self.parameter(
                    'image_orientation_check')),
            inputs={
                'pet_dir': ('pet_recon_dir', directory_format)},
            ouputs={
                'pet_recon_dir_prepared': ('pet_dir_prepared',
                                           directory_format)},
            requirements=[mrtrix_req.v('3.0rc3'), fsl_req.v('5.0.9')])

        return pipeline

    def pet_time_info_extraction_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='pet_info_extraction',
            desc=("Extract PET time info from list-mode header."),
            citations=[],
            **kwargs)

        pipeline.add(
            'PET_time_info',
            PetTimeInfo(),
            inputs={
                'pet_data_dir': ('pet_data_dir', directory_format)},
            ouputs={
                'pet_end_time': ('pet_end_time', float),
                'pet_start_time': ('pet_start_time', str),
                'pet_duration': ('pet_duration', int)})
        return pipeline

    def sinogram_unlisting_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='prepare_sinogram',
            desc=('Unlist pet listmode data into several sinograms and '
                         'perform ssrb compression to prepare data for motion '
                         'detection using PCA pipeline.'),
            citations=[],
            **kwargs)

        prepare_inputs = pipeline.add(
            'prepare_inputs',
            PrepareUnlistingInputs(),
            inputs={
                'list_mode': ('list_mode', list_mode_format),
                'time_offset': ('time_offset', int),
                'num_frames': ('num_frames', int),
                'temporal_len': ('temporal_length', float)})

        unlisting = pipeline.add(
            'unlisting',
            PETListModeUnlisting(),
            inputs={
                'list_inputs': (prepare_inputs, 'out')},
            iterfield=['list_inputs'])

        ssrb = pipeline.add(
            'ssrb',
            SSRB(),
            inputs={
                'unlisted_sinogram': (unlisting, 'pet_sinogram')},
            requirements=[stir_req.v('3.0')])

        pipeline.add(
            'merge_sinograms',
            MergeUnlistingOutputs(),
            inputs={
                'sinograms': (ssrb, 'ssrb_sinograms')},
            ouputs={
                'ssrb_sinograms': ('sinogram_folder', directory_format)},
            joinsource='unlisting',
            joinfield=['sinograms'])

        return pipeline



from .base import PetStudy
from arcana.data import FilesetSpec
from arcana.study.base import StudyMetaClass
from nipype.interfaces.fsl import ExtractROI
from nipype.interfaces.ants.resampling import ApplyTransforms
from arcana.utils.interfaces import Merge
from banana.interfaces.custom.pet import PETdr, GlobalTrendRemoval
from banana.file_format import (nifti_gz_format, png_format,
                                text_matrix_format)
from arcana.study import ParamSpec
import os

template_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__).split('arcana')[0],
                 'arcana', 'reference'))


class DynamicPetStudy(PetStudy, metaclass=StudyMetaClass):

    add_data_specs = [
        FilesetSpec('pet_volumes', nifti_gz_format),
        FilesetSpec('regression_map', nifti_gz_format),
        FilesetSpec('pet_image', nifti_gz_format,
                    'Extract_vol_pipeline'),
        FilesetSpec('registered_volumes', nifti_gz_format,
                    'ApplyTransform_pipeline'),
        FilesetSpec('detrended_volumes', nifti_gz_format,
                    'Baseline_Removal_pipeline'),
        FilesetSpec('spatial_map', nifti_gz_format,
                    'Dual_Regression_pipeline'),
        FilesetSpec('ts', png_format, 'Dual_Regression_pipeline')]

    add_param_specs = [
        ParamSpec('trans_template',
                      os.path.join(template_path, 'PET_template.nii.gz')),
        ParamSpec('base_remove_th', 0),
        ParamSpec('base_remove_binarize', False),
        ParamSpec('regress_th', 0),
        ParamSpec('regress_binarize', False)]

    def Extract_vol_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='Extract_volume',
            desc=('Extract the last volume of the 4D PET timeseries'),
            citations=[],
            **kwargs)

        pipeline.add(
            'fslroi',
            ExtractROI(
                roi_file='vol.nii.gz',
                t_min=79,
                t_size=1),
            inputs={
                'in_file': ('pet_volumes', nifti_gz_format)},
            outputs={
                'pet_image': ('roi_file', nifti_gz_format)})

        return pipeline

    def ApplyTransform_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='applytransform',
            desc=('Apply transformation the the 4D PET timeseries'),
            citations=[],
            **kwargs)

        merge_trans = pipeline.add(
            'merge_transforms',
            Merge(2),
            inputs={
                'in1': ('warp_file', nifti_gz_format),
                'in2': ('affine_mat', text_matrix_format)})

        pipeline.add(
            'ApplyTransform',
            ApplyTransforms(
                reference_image=self.parameter('trans_template'),
                interpolation='Linear',
                input_image_type=3),
            inputs={
                'input_image': ('pet_volumes', nifti_gz_format),
                'transforms': (merge_trans, 'out')},
            outputs={
                'registered_volumes': ('output_image', nifti_gz_format)})

        return pipeline

    def Baseline_Removal_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='Baseline_removal',
            desc=('PET dual regression'),
            citations=[],
            **kwargs)

        pipeline.add(
            'Baseline_removal',
            GlobalTrendRemoval(),
            inputs={
                'volume': ('registered_volumes', nifti_gz_format)},
            outputs={
                'detrended_volumes': ('detrended_file', nifti_gz_format)})

        return pipeline

    def Dual_Regression_pipeline(self, **kwargs):

        pipeline = self.new_pipeline(
            name='Dual_regression',
            desc=('PET dual regression'),
            citations=[],
            **kwargs)

        pipeline.add(
            'PET_dr',
            PETdr(
                threshold=self.parameter('regress_th'),
                binarize=self.parameter('regress_binarize')),
            inputs={
                'volume': ('detrended_volumes', nifti_gz_format),
                'regression_map': ('regression_map', nifti_gz_format)},
            outputs={
                'spatial_map': ('spatial_map', nifti_gz_format),
                'ts': ('timecourse', png_format)})

        return pipeline

    def dynamics_ica_pipeline(self, **kwargs):
        return self._ICA_pipeline_factory(
            input_fileset=FilesetSpec(
                'registered_volumes', nifti_gz_format, **kwargs))



import os
import os.path as op
from banana.exceptions import BananaError


def nth(i):
    "Returns 1st, 2nd, 3rd, 4th, etc for a given number"
    if i == 1:
        s = '1st'
    elif i == 2:
        s = '2nd'
    elif i == 3:
        s = '3rd'
    else:
        s = '{}th'.format(i)
    return s


def get_fsl_reference_path():
    return op.join(os.environ['FSLDIR'], 'data', 'standard')


def get_atlas_path(name, fileset='brain', resolution='1mm'):
    """
    Returns the path to the atlas (or atlas mask) in the arcana repository

    Parameters
    ----------
    name : str
        Name of the Atlas, can be one of ('mni_nl6')
    atlas_type : str
        Whether to return the brain mask or the full atlas, can be one of
        'image', 'mask'
    """
    if name == 'MNI152':
        # MNI ICBM 152 non-linear 6th Generation Symmetric Average Brain
        # Stereotaxic Registration Model (http://nist.mni.mcgill.ca/?p=858)
        if resolution not in ['0.5mm', '1mm', '2mm']:
            raise BananaError(
                "Invalid resolution for MNI152, '{}', can be one of '0.5mm', "
                "'1mm' or '2mm'".format(resolution))
        if fileset == 'image':
            path = op.join(get_fsl_reference_path(),
                                'MNI152_T1_{}.nii.gz'.format(resolution))
        elif fileset == 'mask':
            path = op.join(get_fsl_reference_path(),
                                'MNI152_T1_{}_brain_mask.nii.gz'
                                .format(resolution))
        elif fileset == 'mask_dilated':
            if resolution != '2mm':
                raise BananaError(
                    "Dilated MNI masks are not available for {} resolution "
                    .format(resolution))
            path = op.join(get_fsl_reference_path(),
                                'MNI152_T1_{}_brain_mask_dil.nii.gz'
                                .format(resolution))
        elif fileset == 'brain':
            path = op.join(get_fsl_reference_path(),
                                'MNI152_T1_{}_brain.nii.gz'
                                .format(resolution))
        else:
            raise BananaError("Unrecognised fileset '{}'"
                                  .format(fileset))
    else:
        raise BananaError("Unrecognised atlas name '{}'"
                              .format(name))
    return op.abspath(path)




import os.path
from nipype.interfaces.base import (
    TraitedSpec, BaseInterface, File, Directory, traits, isdefined,
    CommandLineInputSpec, CommandLine)
import pydicom
import nibabel as nib
from arcana.utils import split_extension
import re
from arcana.exceptions import ArcanaError
import numpy as np
from nipype.utils.filemanip import split_filename


class Dcm2niixInputSpec(CommandLineInputSpec):
    input_dir = Directory(mandatory=True, desc='directory name', argstr='"%s"',
                          position=-1)
    compression = traits.Str(argstr='-z %s', desc='type of compression')
    filename = File(genfile=True, argstr='-f %s', desc='output file name')
    out_dir = Directory(genfile=True, argstr='-o %s', desc="output directory")
    multifile_concat = traits.Bool(default=False, desc="concatenate multiple "
                                   "echoes into one file")


class Dcm2niixOutputSpec(TraitedSpec):
    converted = File(exists=True, desc="The converted file")


class Dcm2niix(CommandLine):
    """Convert a DICOM folder to a nifti_gz file"""

    _cmd = 'dcm2niix'
    input_spec = Dcm2niixInputSpec
    output_spec = Dcm2niixOutputSpec

    def _list_outputs(self):
        if (not isdefined(self.inputs.compression) or
                (self.inputs.compression == 'y' or
                 self.inputs.compression == 'i')):
            im_ext = '.nii.gz'
        else:
            im_ext = '.nii'
        outputs = self._outputs().get()
        # As Dcm2niix sometimes prepends a prefix onto the filenames to avoid
        # name clashes with multiple echos, we need to check the output folder
        # for all filenames that end with the "generated filename".
        out_dir = self._gen_filename('out_dir')
        fname = self._gen_filename('filename') + im_ext
        base, ext = split_extension(fname)
        match_re = re.compile(r'(_e\d+)?{}(_(?:e|c)\d+)?{}'
                              .format(base, ext if ext is not None else ''))
        products = [os.path.join(out_dir, f) for f in os.listdir(out_dir)
                    if match_re.match(f) is not None]
        if len(products) == 1:
            converted = products[0]
        elif len(products) > 1 and self.inputs.multifile_concat:
            ex_file = nib.load(products[0])
            data = ex_file.get_data()
            merged_file = np.zeros((data.shape[0], data.shape[1],
                                    data.shape[2], len(products)))
            for i, el in enumerate(products):
                f = nib.load(el)
                merged_file[:, :, :, i] = f.get_data()
            im2save = nib.Nifti1Image(merged_file, ex_file.affine)
            nib.save(im2save, out_dir + fname)
            converted = out_dir + fname
        elif len(products) > 1 and not self.inputs.multifile_concat:
            converted = products[-1]
        else:
            raise ArcanaError("No products produced by dcm2niix ({})"
                                  .format(', '.join(os.listdir(out_dir))))
        outputs['converted'] = converted
        return outputs

    def _gen_filename(self, name):
        if name == 'out_dir':
            fname = self._gen_outdirname()
        elif name == 'filename':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outdirname(self):
        if isdefined(self.inputs.out_dir):
            out_name = self.inputs.out_dir
        else:
            out_name = os.path.join(os.getcwd())
        return out_name

    def _gen_outfilename(self):
        if isdefined(self.inputs.filename):
            out_name = self.inputs.filename
        else:
            out_name = os.path.basename(self.inputs.input_dir)
        return out_name


class Nii2DicomInputSpec(TraitedSpec):
    in_file = File(mandatory=True, desc='input nifti file')
    reference_dicom = traits.List(mandatory=True, desc='original umap')
#     out_file = Directory(genfile=True, desc='the output dicom file')


class Nii2DicomOutputSpec(TraitedSpec):
    out_file = Directory(exists=True, desc='the output dicom file')


class Nii2Dicom(BaseInterface):
    """
    Creates two umaps in dicom format

    fully compatible with the UTE study:

    Attenuation Correction pipeline

    """

    input_spec = Nii2DicomInputSpec
    output_spec = Nii2DicomOutputSpec

    def _run_interface(self, runtime):
        dcms = self.inputs.reference_dicom
        to_remove = [x for x in dcms if '.dcm' not in x]
        if to_remove:
            for f in to_remove:
                dcms.remove(f)
#         dcms = glob.glob(self.inputs.reference_dicom+'/*.dcm')
#         if not dcms:
#             dcms = glob.glob(self.inputs.reference_dicom+'/*.IMA')
#         if not dcms:
#             raise Exception('No DICOM files found in {}'
#                             .format(self.inputs.reference_dicom))
        nifti_image = nib.load(self.inputs.in_file)
        nii_data = nifti_image.get_data()
        if len(dcms) != nii_data.shape[2]:
            raise Exception('Different number of nifti and dicom files '
                            'provided. Dicom to nifti conversion require the '
                            'same number of files in order to run. Please '
                            'check.')
        os.mkdir('nifti2dicom')
        _, basename, _ = split_filename(self.inputs.in_file)
        for i in range(nii_data.shape[2]):
            dcm = pydicom.read_file(dcms[i])
            nifti = nii_data[:, :, i]
            nifti = nifti.astype('uint16')
            dcm.pixel_array.setflags(write=True)
            dcm.pixel_array.flat[:] = nifti.flat[:]
            dcm.PixelData = dcm.pixel_array.T.tostring()
            dcm.save_as('nifti2dicom/{0}_vol{1}.dcm'
                        .format(basename, str(i).zfill(4)))

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = (
            os.getcwd()+'/nifti2dicom')
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            fpath = self.inputs.out_file
        else:
            fname = (
                split_extension(os.path.basename(self.inputs.in_file))[0] +
                '_dicom')
            fpath = os.path.join(os.getcwd(), fname)
        return fpath



import os.path as op
from nipype.interfaces.matlab import MatlabCommand, MatlabInputSpec
from nipype.interfaces.base import TraitedSpec, traits, File

from banana.interfaces import MATLAB_RESOURCES


class BaseMaskInputSpec(MatlabInputSpec):

    # Need to override value in input spec to make it non-mandatory
    script = traits.Str(
        argstr='-r \"%s;exit\"',
        desc='m-code to run',
        position=-1)


class BaseMaskOutputSpec(TraitedSpec):

    raw_output = traits.Str("Raw output of the matlab command")


class BaseMask(MatlabCommand):
    """
    Base class for MATLAB mask interfaces
    """

    def run(self, **inputs):
        self.work_dir = inputs['cwd']
        # Set the script input of the matlab spec
        self.inputs.script = (
            "set_param(0,'CharacterEncoding','UTF-8');\n"
            "addpath(genpath('{}'));\n".format(MATLAB_RESOURCES) +
            self.script(**inputs) +
            "exit;")
        results = super().run(**inputs)
        stdout = results.runtime.stdout
        # Attach stdout to outputs to access matlab results
        results.outputs.raw_output = stdout
        return results

    def script(self, **inputs):  # @UnusedVariable
        """
        Generate script to perform masking
        """
        raise NotImplementedError

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs['out_file'] = self.out_file
        return outputs

    @property
    def out_file(self):
        return op.realpath(op.abspath(op.join(self.work_dir, 'out_file.nii')))


class DialateMaskInputSpec(BaseMaskInputSpec):

    in_file = File(exists=True, desc="Mask to dialate")
    dialation = traits.List((traits.Float(), traits.Float(), traits.Float),
                            desc="Size of the dialation")


class DialateMaskOutputSpec(BaseMaskOutputSpec):

    out_file = File(exists=True, desc="Dialated mask")


class DialateMask(BaseMask):
    """
    Base interface for STI classes
    """
    input_spec = DialateMaskInputSpec
    output_spec = DialateMaskOutputSpec

    def script(self, **inputs):  # @UnusedVariable
        """
        Generate script to load images, pass them to the STI function along
        with the keyword parameters
        """
        script = """
            % Spherical structure element
            SE = fspecial3('ellipsoid', {dialation});

            % Load whole brain mask
            mask = load_untouch_nii('{in_file}');
            mask_array = mask.img > 0;
            dialated = imdilate(mask_array, SE > 0);
            mask.img = dialated
            save_untouch_nii(mask, '{out_file}');
        """.format(
            dialation=self.inputs.dialation,
            in_file=self.inputs.in_file,
            out_file=self.out_file)
        return script


class MaskCoilsInputSpec(BaseMaskInputSpec):

    masks = traits.List(File(mandatory=True), desc="Input masks")

    dialation = traits.List((traits.Float(), traits.Float(), traits.Float),
                            desc="Size of the dialation")
    whole_brain_mask = File(mandatory=True, desc="Whole brain mask")


class MaskCoilsOutputSpec(BaseMaskOutputSpec):

    out_files = traits.List(File(exists=True), desc="Output files")


class MaskCoils(BaseMask):
    """
    Generate coil specific masks by thresholding magnitude image
    """

    input_spec = MaskCoilsInputSpec
    output_spec = MaskCoilsOutputSpec

    def script(self, **inputs):  # @UnusedVariable
        """
        Generate script to load images, pass them to the STI function along
        with the keyword parameters
        """
        script = """
            % Spherical structure element
            SE = fspecial3('ellipsoid', [11 11 11]);

            % List all input files
            mask_files = {{'{masks}'}}

            % Load whole brain mask
            whole_brain_mask = load_untouch_nii('{whole_brain_mask}');

            for i=1:length(mask_files)
                mag = load_untouch_nii(mask_files{{i}});

                % Blur to remove tissue based contrast
                vol = convn(mag.img, SE, 'same');
                % Threshold to high-signal area
                vol = vol > graythresh(mag.img);
                % Remove orphaned pixels and then close holes in WB mask
                vol = imclose(vol, SE > 0) > 0;
                vol = imopen(vol, SE > 0) > 0;

                % Clip to brain whole_brain_mask region
                mask = mag;
                mask.img = (vol .* whole_brain_mask.img) > 0;

                save_untouch_nii(mask, ['{out_file_base}' num2str(i) '.nii']);
            end
        """.format(
            masks="', '".join(self.inputs.masks),
            whole_brain_mask=self.inputs.whole_brain_mask,
            out_file_base=self.out_file_base)
        return script

    def _list_outputs(self):
        outputs = self._outputs().get()
        base = self.out_file_base
        outputs['out_files'] = ['{}{}.nii'.format(base, i)
                                for i in range(1, len(self.inputs.masks) + 1)]
        return outputs

    @property
    def out_file_base(self):
        return op.realpath(op.abspath(op.join(self.work_dir, 'out_file')))


class MedianInMasksInputSpec(BaseMaskInputSpec):

    channels = traits.List(File(), mandatory=True,
                           desc="Input mask to dialate")
    channel_masks = traits.List(File(), mandatory=True,
                                desc="Separate masks for each input file")
    whole_brain_mask = File(mandatory=True, desc="Whole brain mask")
    out_file = File(genfile=True, desc="Name of the output file")
    # Need to override value in input spec to make it non-mandatory


class MedianInMasksOutputSpec(BaseMaskOutputSpec):

    out_file = File(exists=True, desc="Output file")


class MedianInMasks(BaseMask):
    """
    Calculate the median value between voxels across the provided images
    that are not masked out.
    """
    input_spec = MedianInMasksInputSpec
    output_spec = MedianInMasksOutputSpec

    def script(self, **inputs):  # @UnusedVariable
        """
        Generate script to load images, pass them to the STI function along
        with the keyword parameters
        """
        script = """
            brainMask = load_untouch_nii('{mask}');
            brainMask = brainMask.img>0;

            qsmVol = [];
            missingValues = [];
            dims = [];

            channel_files = {{{channel_files}}};
            mask_files = {{{mask_files}}};

            num_channels = length(channel_files);

            for i=1:length(channel_files)
                nii = load_untouch_nii(channel_files{{i}});
                mask = load_untouch_nii(mask_files{{i}});

                qsmVol(:,i) = nii.img(:).*(mask.img(:)>0)-99*(mask.img(:)==0);
                missingValues(:,i) = mask.img(:)==0;

                if isempty(dims)
                    dims = size(nii.img);
                end
            end

            % Order values so median value is at index 16
            qsmVol = sort(qsmVol,2);

            % Adjust median index (16) based on missing values
            indVol=sub2ind(size(qsmVol),1:size(qsmVol,1),num_channels-floor(0.5*(num_channels-sum(missingValues,2)')));

            % Take median value using index, resize and mask out background
            medVol=reshape(qsmVol(indVol),dims);
            medVol(medVol==-99) = 0;
            medVol(brainMask==0) = 0;

            % Save output
            nii.img = medVol;
            save_untouch_nii(nii,'{out_file}');
        """.format(
            channel_files=','.join(
                "'{}'".format(f) for f in self.inputs.channels),
            mask_files=','.join(
                "'{}'".format(f) for f in self.inputs.channel_masks),
            mask=self.inputs.whole_brain_mask,
            out_file=self.out_file)
        return script



from logging import getLogger
from nipype.interfaces.utility import Merge
from nipype.interfaces.mrtrix3 import ResponseSD, Tractography
from nipype.interfaces.mrtrix3.utils import BrainMask, TensorMetrics
from nipype.interfaces.mrtrix3.reconst import FitTensor, EstimateFOD
from banana.interfaces.mrtrix import (
    DWIPreproc, MRCat, ExtractDWIorB0, MRMath, DWIBiasCorrect, DWIDenoise,
    MRCalc, DWIIntensityNorm, AverageResponse, DWI2Mask)
# from nipype.workflows.dwi.fsl.tbss import create_tbss_all
# from banana.interfaces.noddi import (
#     CreateROI, BatchNODDIFitting, SaveParamsAsNIfTI)
from banana.interfaces.mrtrix import MRConvert, ExtractFSLGradients
from arcana.utils.interfaces import MergeTuple, Chain
from nipype.interfaces.utility import IdentityInterface
from banana.citation import (
    mrtrix_cite, fsl_cite, eddy_cite, topup_cite, distort_correct_cite,
    fast_cite, n4_cite, dwidenoise_cites)
from banana.file_format import (
    mrtrix_image_format, nifti_gz_format, nifti_gz_x_format, fsl_bvecs_format,
    fsl_bvals_format, text_format, dicom_format, eddy_par_format,
    mrtrix_track_format, motion_mats_format)
from banana.requirement import (
    fsl_req, mrtrix_req, ants_req)
from arcana.data import FilesetSpec, FilesetInputSpec
from arcana.utils.interfaces import SelectSession
from arcana.study import ParamSpec, SwitchSpec
from .epi import EpiStudy
from nipype.interfaces import fsl
from banana.interfaces.custom.motion_correction import (
    PrepareDWI, AffineMatrixGeneration)
from banana.study.base import Study
from banana.bids import BidsInput, BidsAssocInput
from banana.exceptions import BananaUsageError
from arcana.exceptions import ArcanaMissingDataException
from banana.study import StudyMetaClass

logger = getLogger('banana')


class DwiStudy(EpiStudy, metaclass=StudyMetaClass):

    add_data_specs = [
        FilesetInputSpec('dwi_reference', nifti_gz_format, optional=True),
        FilesetSpec('b0', nifti_gz_format, 'extract_b0_pipeline',
                    desc="b0 image"),
        FilesetSpec('noise_residual', mrtrix_image_format, 'preprocess_pipeline'),
        FilesetSpec('tensor', nifti_gz_format, 'tensor_pipeline'),
        FilesetSpec('fa', nifti_gz_format, 'tensor_metrics_pipeline'),
        FilesetSpec('adc', nifti_gz_format, 'tensor_metrics_pipeline'),
        FilesetSpec('wm_response', text_format, 'response_pipeline'),
        FilesetSpec('gm_response', text_format, 'response_pipeline'),
        FilesetSpec('csf_response', text_format, 'response_pipeline'),
        FilesetSpec('avg_response', text_format,
                    'average_response_pipeline'),
        FilesetSpec('wm_odf', mrtrix_image_format, 'fod_pipeline'),
        FilesetSpec('gm_odf', mrtrix_image_format, 'fod_pipeline'),
        FilesetSpec('csf_odf', mrtrix_image_format, 'fod_pipeline'),
        FilesetSpec('bias_correct', nifti_gz_format,
                    'bias_correct_pipeline'),
        FilesetSpec('grad_dirs', fsl_bvecs_format, 'preprocess_pipeline'),
        FilesetSpec('bvalues', fsl_bvals_format, 'preprocess_pipeline'),
        FilesetSpec('eddy_par', eddy_par_format, 'preprocess_pipeline'),
        FilesetSpec('brain', nifti_gz_format,
                    'brain_extraction_pipeline'),
        FilesetSpec('brain_mask', nifti_gz_format,
                    'brain_extraction_pipeline'),
        FilesetSpec('norm_intensity', mrtrix_image_format,
                    'intensity_normalisation_pipeline'),
        FilesetSpec('norm_intens_fa_template', mrtrix_image_format,
                    'intensity_normalisation_pipeline',
                    frequency='per_study'),
        FilesetSpec('norm_intens_wm_mask', mrtrix_image_format,
                    'intensity_normalisation_pipeline',
                    frequency='per_study'),
        FilesetSpec('global_tracks', mrtrix_track_format,
                    'global_tracking_pipeline'),
        FilesetSpec('wm_mask', mrtrix_image_format,
                    'global_tracking_pipeline')
        # FilesetSpec('tbss_mean_fa', nifti_gz_format, 'tbss_pipeline',
        #             frequency='per_study'),
        # FilesetSpec('tbss_proj_fa', nifti_gz_format, 'tbss_pipeline',
        #             frequency='per_study'),
        # FilesetSpec('tbss_skeleton', nifti_gz_format, 'tbss_pipeline',
        #             frequency='per_study'),
        # FilesetSpec('tbss_skeleton_mask', nifti_gz_format,
        #             'tbss_pipeline', frequency='per_study'),
        ]  # @IgnorePep8

    add_param_specs = [
        ParamSpec('multi_tissue', True),
        ParamSpec('preproc_pe_dir', None, dtype=str),
        ParamSpec('tbss_skel_thresh', 0.2),
        ParamSpec('fsl_mask_f', 0.25),
        ParamSpec('bet_robust', True),
        ParamSpec('bet_f_threshold', 0.2),
        ParamSpec('bet_reduce_bias', False),
        ParamSpec('num_global_tracks', int(1e9)),
        ParamSpec('global_tracks_cutoff', 0.05),
        SwitchSpec('preproc_denoise', False),
        SwitchSpec('response_algorithm', 'tax',
                   ('tax', 'dhollander', 'msmt_5tt')),
        SwitchSpec('fod_algorithm', 'csd', ('csd', 'msmt_csd')),
        SwitchSpec('brain_extract_method', 'mrtrix',
                   ('mrtrix', 'fsl')),
        SwitchSpec('bias_correct_method', 'ants', ('ants', 'fsl'))]

    primary_bids_selector = BidsInput(
        spec_name='magnitude', type='dwi', format=nifti_gz_x_format)

    default_bids_inputs = [primary_bids_selector,
                           BidsAssocInput(
                               spec_name='bvalues',
                               primary=primary_bids_selector,
                               association='grads',
                               type='bval',
                               format=fsl_bvals_format),
                           BidsAssocInput(
                               spec_name='grad_dirs',
                               primary=primary_bids_selector,
                               association='grads',
                               type='bvec',
                               format=fsl_bvecs_format),
                           BidsAssocInput(
                               spec_name='reverse_phase',
                               primary=primary_bids_selector,
                               association='epi',
                               format=nifti_gz_format,
                               drop_if_missing=True)]

    @property
    def multi_tissue(self):
        return self.branch('response_algorithm',
                           ('msmt_5tt', 'dhollander'))

    def preprocess_pipeline(self, **name_maps):
        """
        Performs a series of FSL preprocessing steps, including Eddy and Topup

        Parameters
        ----------
        phase_dir : str{AP|LR|IS}
            The phase encode direction
        """

        # Determine whether we can correct for distortion, i.e. if reference
        # scans are provided
        # Include all references
        references = [fsl_cite, eddy_cite, topup_cite,
                      distort_correct_cite]
        if self.branch('preproc_denoise'):
            references.extend(dwidenoise_cites)

        pipeline = self.new_pipeline(
            name='preprocess',
            name_maps=name_maps,
            desc=(
                "Preprocess dMRI studies using distortion correction"),
            citations=references)

        # Create nodes to gradients to FSL format
        if self.input('magnitude').format == dicom_format:
            extract_grad = pipeline.add(
                "extract_grad",
                ExtractFSLGradients(),
                inputs={
                    'in_file': ('magnitude', dicom_format)},
                outputs={
                    'grad_dirs': ('bvecs_file', fsl_bvecs_format),
                    'bvalues': ('bvals_file', fsl_bvals_format)},
                requirements=[mrtrix_req.v('3.0rc3')])
            grad_fsl_inputs = {'in1': (extract_grad, 'bvecs_file'),
                               'in2': (extract_grad, 'bvals_file')}
        elif self.provided('grad_dirs') and self.provided('bvalues'):
            grad_fsl_inputs = {'in1': ('grad_dirs', fsl_bvecs_format),
                               'in2': ('bvalues', fsl_bvals_format)}
        else:
            raise BananaUsageError(
                "Either input 'magnitude' image needs to be in DICOM format "
                "or gradient directions and b-values need to be explicitly "
                "provided to {}".format(self))

        # Gradient merge node
        grad_fsl = pipeline.add(
            "grad_fsl",
            MergeTuple(2),
            inputs=grad_fsl_inputs)

        # Denoise the dwi-scan
        if self.branch('preproc_denoise'):
            # Run denoising
            denoise = pipeline.add(
                'denoise',
                DWIDenoise(),
                inputs={
                    'in_file': ('magnitude', nifti_gz_format)},
                requirements=[mrtrix_req.v('3.0rc3')])

            # Calculate residual noise
            subtract_operands = pipeline.add(
                'subtract_operands',
                Merge(2),
                inputs={
                    'in1': ('magnitude', nifti_gz_format),
                    'in2': (denoise, 'noise')})

            pipeline.add(
                'subtract',
                MRCalc(
                    operation='subtract'),
                inputs={
                    'operands': (subtract_operands, 'out')},
                outputs={
                    'noise_residual': ('out_file', mrtrix_image_format)},
                requirements=[mrtrix_req.v('3.0rc3')])

        # Preproc kwargs
        preproc_kwargs = {}
        preproc_inputs = {'grad_fsl': (grad_fsl, 'out')}

        if (self.provided('dwi_reference') or
                self.provided('reverse_phase')):
            # Extract b=0 volumes
            dwiextract = pipeline.add(
                'dwiextract',
                ExtractDWIorB0(
                    bzero=True,
                    out_ext='.nii.gz'),
                inputs={
                    'in_file': ('magnitude', dicom_format)},
                requirements=[mrtrix_req.v('3.0rc3')])

            # Get first b=0 from dwi b=0 volumes
            mrconvert = pipeline.add(
                "mrconvert",
                MRConvert(
                    coord=(3, 0)),
                inputs={
                    'in_file': (dwiextract, 'out_file')},
                requirements=[mrtrix_req.v('3.0rc3')])

            # Concatenate extracted forward rpe with reverse rpe
            mrcat = pipeline.add(
                'mrcat',
                MRCat(),
                inputs={
                    'second_scan': ((
                        'dwi_reference' if self.provided('dwi_reference')
                        else 'reverse_phase'), mrtrix_image_format),
                    'first_scan': (mrconvert, 'out_file')},
                requirements=[mrtrix_req.v('3.0rc3')])

            # Create node to assign the right PED to the diffusion
            prep_dwi = pipeline.add(
                'prepare_dwi',
                PrepareDWI(),
                inputs={
                    'pe_dir': ('ped', float),
                    'ped_polarity': ('pe_angle', float)})

            preproc_kwargs['rpe_pair'] = True

            distortion_correction = True
            preproc_inputs['se_epi'] = (mrcat, 'out_file')
        else:
            distortion_correction = False
            preproc_kwargs['rpe_none'] = True

        if self.parameter('preproc_pe_dir') is not None:
            preproc_kwargs['pe_dir'] = self.parameter('preproc_pe_dir')

        preproc = pipeline.add(
            'dwipreproc',
            DWIPreproc(
                no_clean_up=True,
                out_file_ext='.nii.gz',
                # FIXME: Need to determine this programmatically
                # eddy_parameters = '--data_is_shelled '
                temp_dir='dwipreproc_tempdir',
                **preproc_kwargs),
            inputs=preproc_inputs,
            outputs={
                'eddy_par': ('eddy_parameters', eddy_par_format)},
            requirements=[mrtrix_req.v('3.0rc3'), fsl_req.v('5.0.10')],
            wall_time=60)
        if self.branch('preproc_denoise'):
            pipeline.connect(denoise, 'out_file', preproc, 'in_file')
        else:
            pipeline.connect_input('magnitude', preproc, 'in_file',
                                   nifti_gz_format)
        if distortion_correction:
            pipeline.connect(prep_dwi, 'pe', preproc, 'pe_dir')

        # Create node to reorient preproc out_file
        pipeline.add(
            'fslreorient2std',
            fsl.utils.Reorient2Std(),
            inputs={
                'in_file': (preproc, 'out_file')},
            outputs={
                'preproc': ('out_file', nifti_gz_format)},
            requirements=[fsl_req.v('5.0.9')])

        return pipeline

    def brain_extraction_pipeline(self, **name_maps):
        """
        Generates a whole brain mask using MRtrix's 'dwi2mask' command

        Parameters
        ----------
        mask_tool: Str
            Can be either 'bet' or 'dwi2mask' depending on which mask tool you
            want to use
        """

        if self.branch('brain_extract_method', 'mrtrix'):
            pipeline = self.new_pipeline(
                'brain_extraction',
                desc="Generate brain mask from b0 images",
                citations=[mrtrix_cite],
                name_maps=name_maps)

            # Gradient merge node
            grad_fsl = pipeline.add(
                "grad_fsl",
                MergeTuple(2),
                inputs={
                    'in1': ('grad_dirs', fsl_bvecs_format),
                    'in2': ('bvalues', fsl_bvals_format)})

            # Create mask node
            pipeline.add(
                'dwi2mask',
                BrainMask(
                    out_file='brain_mask.nii.gz'),
                inputs={
                    'in_file': ('preproc', nifti_gz_format),
                    'grad_fsl': (grad_fsl, 'out')},
                outputs={
                    'brain_mask': ('out_file', nifti_gz_format)},
                requirements=[mrtrix_req.v('3.0rc3')])

        else:
            pipeline = super(DwiStudy, self).brain_extraction_pipeline(
                **name_maps)
        return pipeline

    def bias_correct_pipeline(self, **name_maps):
        """
        Corrects B1 field inhomogeneities
        """

        pipeline = self.new_pipeline(
            name='bias_correct',
            desc="Corrects for B1 field inhomogeneity",
            citations=[fast_cite,
                        (n4_cite
                         if self.parameter('bias_correct_method') == 'ants'
                         else fsl_cite)],
            name_maps=name_maps)

        # Gradient merge node
        fsl_grads = pipeline.add(
            "fsl_grads",
            MergeTuple(2),
            inputs={
                'in1': ('grad_dirs', fsl_bvecs_format),
                'in2': ('bvalues', fsl_bvals_format)})

        # Create bias correct node
        pipeline.add(
            "bias_correct",
            DWIBiasCorrect(
                method=self.parameter('bias_correct_method')),
            inputs={
                'grad_fsl': (fsl_grads, 'out'),  # internal
                'in_file': ('preproc', nifti_gz_format),
                'mask': ('brain_mask', nifti_gz_format)},
            outputs={
                'bias_correct': ('out_file', nifti_gz_format)},
            requirements=(
                [mrtrix_req.v('3.0rc3'),
                 (ants_req.v('2.0')
                  if self.parameter('bias_correct_method') == 'ants'
                  else fsl_req.v('5.0.9'))]))

        return pipeline

    def intensity_normalisation_pipeline(self, **name_maps):

        if self.num_sessions < 2:
            raise ArcanaMissingDataException(
                "Cannot normalise intensities of DWI images as study only "
                "contains a single session")
        elif self.num_sessions < self.RECOMMENDED_NUM_SESSIONS_FOR_INTENS_NORM:
            logger.warning(
                "The number of sessions in the study ({}) is less than the "
                "recommended number for intensity normalisation ({}). The "
                "results may be unreliable".format(
                    self.num_sessions,
                    self.RECOMMENDED_NUM_SESSIONS_FOR_INTENS_NORM))

        pipeline = self.new_pipeline(
            name='intensity_normalization',
            desc="Corrects for B1 field inhomogeneity",
            citations=[mrtrix_req.v('3.0rc3')],
            name_maps=name_maps)

        # Convert from nifti to mrtrix format
        grad_merge = pipeline.add(
            "grad_merge",
            MergeTuple(2),
            inputs={
                'in1': ('grad_dirs', fsl_bvecs_format),
                'in2': ('bvalues', fsl_bvals_format)})

        mrconvert = pipeline.add(
            'mrconvert',
            MRConvert(
                out_ext='.mif'),
            inputs={
                'in_file': ('bias_correct', nifti_gz_format),
                'grad_fsl': (grad_merge, 'out')})

        # Set up join nodes
        fields = ['dwis', 'masks', 'subject_ids', 'visit_ids']
        join_subjects = pipeline.add(
            'join_subjects',
            IdentityInterface(
                fields),
            inputs={
                'masks': ('brain_mask', nifti_gz_format),
                'dwis': (mrconvert, 'out_file'),
                'subject_ids': (Study.SUBJECT_ID, int),
                'visit_ids': (Study.VISIT_ID, int)},
            joinsource=self.SUBJECT_ID,
            joinfield=fields)

        join_visits = pipeline.add(
            'join_visits',
            Chain(
                fields),
            inputs={
                'dwis': (join_subjects, 'dwis'),
                'masks': (join_subjects, 'masks'),
                'subject_ids': (join_subjects, 'subject_ids'),
                'visit_ids': (join_subjects, 'visit_ids')},
            joinsource=self.VISIT_ID,
            joinfield=fields)

        # Intensity normalization
        intensity_norm = pipeline.add(
            'dwiintensitynorm',
            DWIIntensityNorm(),
            inputs={
                'in_files': (join_visits, 'dwis'),
                'masks': (join_visits, 'masks')},
            outputs={
                'norm_intens_fa_template': ('fa_template', mrtrix_image_format),
                'norm_intens_wm_mask': ('wm_mask', mrtrix_image_format)})

        # Set up expand nodes
        pipeline.add(
            'expand', SelectSession(),
            inputs={
                'subject_ids': (join_visits, 'subject_ids'),
                'visit_ids': (join_visits, 'visit_ids'),
                'items': (intensity_norm, 'out_files'),
                'subject_id': (Study.SUBJECT_ID, int),
                'visit_id': (Study.VISIT_ID, int)},
            outputs={
                'norm_intensity': ('item', mrtrix_image_format)})

        # Connect inputs
        return pipeline

    def tensor_pipeline(self, **name_maps):  # @UnusedVariable
        """
        Fits the apparrent diffusion tensor (DT) to each voxel of the image
        """

        pipeline = self.new_pipeline(
            name='tensor',
            desc=("Estimates the apparent diffusion tensor in each "
                  "voxel"),
            citations=[],
            name_maps=name_maps)

        # Gradient merge node
        fsl_grads = pipeline.add(
            "fsl_grads",
            MergeTuple(2),
            inputs={
                'in1': ('grad_dirs', fsl_bvecs_format),
                'in2': ('bvalues', fsl_bvals_format)})

        # Create tensor fit node
        pipeline.add(
            'dwi2tensor',
            FitTensor(
                out_file='dti.nii.gz'),
            inputs={
                'grad_fsl': (fsl_grads, 'out'),
                'in_file': ('bias_correct', nifti_gz_format),
                'in_mask': ('brain_mask', nifti_gz_format)},
            outputs={
                'tensor': ('out_file', nifti_gz_format)})

        return pipeline

    def tensor_metrics_pipeline(self, **name_maps):  # @UnusedVariable
        """
        Fits the apparrent diffusion tensor (DT) to each voxel of the image
        """

        pipeline = self.new_pipeline(
            name='fa',
            desc=("Calculates the FA and ADC from a tensor image"),
            citations=[],
            name_maps=name_maps)

        # Create tensor fit node
        pipeline.add(
            'metrics',
            TensorMetrics(
                out_fa='fa.nii.gz',
                out_adc='adc.nii.gz'),
            inputs={
                'in_file': ('tensor', nifti_gz_format),
                'in_mask': ('brain_mask', nifti_gz_format)},
            outputs={
                'fa': ('out_fa', nifti_gz_format),
                'adc': ('out_adc', nifti_gz_format)},
            requirements=[mrtrix_req.v('3.0rc3')])

        return pipeline

    def response_pipeline(self, **name_maps):  # @UnusedVariable
        """
        Estimates the fibre orientation distribution (FOD) using constrained
        spherical deconvolution

        Parameters
        ----------
        response_algorithm : str
            Algorithm used to estimate the response
        """

        pipeline = self.new_pipeline(
            name='response',
            desc=("Estimates the fibre response function"),
            citations=[mrtrix_cite],
            name_maps=name_maps)

        # Gradient merge node
        fsl_grads = pipeline.add(
            "fsl_grads",
            MergeTuple(2),
            inputs={
                'in1': ('grad_dirs', fsl_bvecs_format),
                'in2': ('bvalues', fsl_bvals_format)})

        # Create fod fit node
        response = pipeline.add(
            'response',
            ResponseSD(
                algorithm=self.parameter('response_algorithm')),
            inputs={
                'grad_fsl': (fsl_grads, 'out'),
                'in_file': ('bias_correct', nifti_gz_format),
                'in_mask': ('brain_mask', nifti_gz_format)},
            outputs={
                'wm_response': ('wm_file', text_format)},
            requirements=[mrtrix_req.v('3.0rc3')])

        # Connect to outputs
        if self.multi_tissue:
            response.inputs.gm_file = 'gm.txt',
            response.inputs.csf_file = 'csf.txt',
            pipeline.connect_output('gm_response', response, 'gm_file',
                                    text_format)
            pipeline.connect_output('csf_response', response, 'csf_file',
                                    text_format)

        return pipeline

    def average_response_pipeline(self, **name_maps):
        """
        Averages the estimate response function over all subjects in the
        project
        """

        pipeline = self.new_pipeline(
            name='average_response',
            desc=(
                "Averages the fibre response function over the project"),
            citations=[mrtrix_cite],
            name_maps=name_maps)

        join_subjects = pipeline.add(
            'join_subjects',
            IdentityInterface(['responses']),
            inputs={
                'responses': ('wm_response', text_format)},
            outputs={},
            joinsource=self.SUBJECT_ID,
            joinfield=['responses'])

        join_visits = pipeline.add(
            'join_visits',
            Chain(['responses']),
            inputs={
                'responses': (join_subjects, 'responses')},
            joinsource=self.VISIT_ID,
            joinfield=['responses'])

        pipeline.add(
            'avg_response',
            AverageResponse(),
            inputs={
                'in_files': (join_visits, 'responses')},
            outputs={
                'avg_response': ('out_file', text_format)})

        return pipeline

    def fod_pipeline(self, **name_maps):  # @UnusedVariable
        """
        Estimates the fibre orientation distribution (FOD) using constrained
        spherical deconvolution

        Parameters
        ----------
        """

        pipeline = self.new_pipeline(
            name='fod',
            desc=("Estimates the fibre orientation distribution in each"
                  " voxel"),
            citations=[mrtrix_cite],
            name_maps=name_maps)

        if self.branch('fod_algorithm', 'msmt_csd'):
            pipeline.add_input(FilesetSpec('gm_response', text_format))
            pipeline.add_input(FilesetSpec('csf_response', text_format))

        # Gradient merge node
        fsl_grads = pipeline.add(
            "fsl_grads",
            MergeTuple(2),
            inputs={
                'in1': ('grad_dirs', fsl_bvecs_format),
                'in2': ('bvalues', fsl_bvals_format)})

        # Create fod fit node
        dwi2fod = pipeline.add(
            'dwi2fod',
            EstimateFOD(
                algorithm=self.parameter('fod_algorithm')),
            inputs={
                'in_file': ('bias_correct', nifti_gz_format),
                'wm_txt': ('wm_response', text_format),
                'mask_file': ('brain_mask', nifti_gz_format),
                'grad_fsl': (fsl_grads, 'out')},
            outputs={
                'wm_odf': ('wm_odf', nifti_gz_format)},
            requirements=[mrtrix_req.v('3.0rc3')])

        if self.multi_tissue:
            dwi2fod.inputs.gm_odf = 'gm.mif',
            dwi2fod.inputs.csf_odf = 'csf.mif',
            pipeline.connect_input('gm_response', dwi2fod, 'gm_txt',
                                   text_format),
            pipeline.connect_input('csf_response', dwi2fod, 'csf_txt',
                                   text_format),
            pipeline.connect_output('gm_odf', dwi2fod, 'gm_odf',
                                    nifti_gz_format),
            pipeline.connect_output('csf_odf', dwi2fod, 'csf_odf',
                                    nifti_gz_format),
        # Check inputs/output are connected
        return pipeline

    def extract_b0_pipeline(self, **name_maps):  # @UnusedVariable
        """
        Extracts the b0 images from a DWI study and takes their mean
        """

        pipeline = self.new_pipeline(
            name='extract_b0',
            desc="Extract b0 image from a DWI study",
            citations=[mrtrix_cite],
            name_maps=name_maps)

        # Gradient merge node
        fsl_grads = pipeline.add(
            "fsl_grads",
            MergeTuple(2),
            inputs={
                'in1': ('grad_dirs', fsl_bvecs_format),
                'in2': ('bvalues', fsl_bvals_format)})

        # Extraction node
        extract_b0s = pipeline.add(
            'extract_b0s',
            ExtractDWIorB0(
                bzero=True,
                quiet=True),
            inputs={
                'fslgrad': (fsl_grads, 'out'),
                'in_file': ('bias_correct', nifti_gz_format)},
            requirements=[mrtrix_req.v('3.0rc3')])

        # FIXME: Need a registration step before the mean
        # Mean calculation node
        mean = pipeline.add(
            "mean",
            MRMath(
                axis=3,
                operation='mean',
                quiet=True),
            inputs={
                'in_files': (extract_b0s, 'out_file')},
            requirements=[mrtrix_req.v('3.0rc3')])

        # Convert to Nifti
        pipeline.add(
            "output_conversion",
            MRConvert(
                out_ext='.nii.gz',
                quiet=True),
            inputs={
                'in_file': (mean, 'out_file')},
            outputs={
                'b0': ('out_file', nifti_gz_format)},
            requirements=[mrtrix_req.v('3.0rc3')])

        return pipeline

    def global_tracking_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='global_tracking',
            desc="Extract b0 image from a DWI study",
            citations=[mrtrix_cite],
            name_maps=name_maps)

        # Add gradients to input image
        fsl_grads = pipeline.add(
            "fsl_grads",
            MergeTuple(2),
            inputs={
                'in1': ('grad_dirs', fsl_bvecs_format),
                'in2': ('bvalues', fsl_bvals_format)})

        mask = pipeline.add(
            'mask',
            DWI2Mask(),
            inputs={
                'grad_fsl': (fsl_grads, 'out'),
                'in_file': ('bias_correct', nifti_gz_format)})

        pipeline.add(
            'tracking',
            Tractography(
                select=self.parameter('num_global_tracks'),
                cutoff=self.parameter('global_tracks_cutoff')),
            inputs={
                'seed_image': (mask, 'out_file'),
                'in_file': ('wm_odf', mrtrix_image_format)},
            outputs={
                'global_tracks': ('out_file', mrtrix_track_format)})

        return pipeline

    def intrascan_alignment_pipeline(self, **name_maps):

        pipeline = self.new_pipeline(
            name='affine_mat_generation',
            desc=("Generation of the affine matrices for the main dwi "
                  "sequence starting from eddy motion parameters"),
            citations=[fsl_cite],
            name_maps=name_maps)

        pipeline.add(
            'gen_aff_mats',
            AffineMatrixGeneration(),
            inputs={
                'reference_image': ('preproc', nifti_gz_format),
                'motion_parameters': ('eddy_par', eddy_par_format)},
            outputs={
                'align_mats': ('affine_matrices', motion_mats_format)})

        return pipeline

#     def tbss_pipeline(self, **name_maps):  # @UnusedVariable
#
# #             inputs=[FilesetSpec('fa', nifti_gz_format)],
# #             outputs=[FilesetSpec('tbss_mean_fa', nifti_gz_format),
# #                      FilesetSpec('tbss_proj_fa', nifti_gz_format,
# #                                  frequency='per_study'),
# #                      FilesetSpec('tbss_skeleton', nifti_gz_format,
# #                                  frequency='per_study'),
# #                      FilesetSpec('tbss_skeleton_mask', nifti_gz_format,
# #                                  frequency='per_study')],
#         pipeline = self.new_pipeline(
#             name='tbss',
#             citations=[tbss_cite, fsl_cite],
#             name_maps=name_maps)
#         # Create TBSS workflow
# #         tbss = create_tbss_all(name='tbss')
#         # Connect inputs
#         pipeline.connect_input('fa', tbss, 'inputnode.fa_list')
#         # Connect outputs
#         pipeline.connect_output('tbss_mean_fa', tbss,
#                                 'outputnode.meanfa_file')
#         pipeline.connect_output('tbss_proj_fa', tbss,
#                                 'outputnode.projectedfa_file')
#         pipeline.connect_output('tbss_skeleton', tbss,
#                                 'outputnode.skeleton_file')
#         pipeline.connect_output('tbss_skeleton_mask', tbss,
#                                 'outputnode.skeleton_mask')
#         # Check inputs/output are connected
#         return pipeline


# class NODDIStudy(DwiStudy, metaclass=StudyMetaClass):
# 
#     add_data_specs = [
#         FilesetInputSpec('low_b_dw_scan', mrtrix_image_format),
#         FilesetInputSpec('high_b_dw_scan', mrtrix_image_format),
#         FilesetSpec('dwi_scan', mrtrix_image_format, 'concatenate_pipeline'),
#         FilesetSpec('ficvf', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('odi', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('fiso', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('fibredirs_xvec', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('fibredirs_yvec', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('fibredirs_zvec', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('fmin', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('kappa', nifti_format, 'noddi_fitting_pipeline'),
#         FilesetSpec('error_code', nifti_format, 'noddi_fitting_pipeline')]
# 
#     add_param_specs = [ParamSpec('noddi_model',
#                                          'WatsonSHStickTortIsoV_B0'),
#                            SwitchSpec('single_slice', False)]
# 
#     def concatenate_pipeline(self, **name_maps):  # @UnusedVariable
#         """
#         Concatenates two dMRI filesets (with different b-values) along the
#         DW encoding (4th) axis
#         """
# #             inputs=[FilesetSpec('low_b_dw_scan', mrtrix_image_format),
# #                     FilesetSpec('high_b_dw_scan', mrtrix_image_format)],
# #             outputs=[FilesetSpec('dwi_scan', mrtrix_image_format)],
#         pipeline = self.new_pipeline(
#             name='concatenation',
# 
#             desc=(
#                 "Concatenate low and high b-value dMRI filesets for NODDI "
#                 "processing"),
#             citations=[mrtrix_cite],
#             name_maps=name_maps)
#         # Create concatenation node
#         mrcat = pipeline.add('mrcat', MRCat(),
#                                      requirements=[mrtrix_req.v('3.0rc3')])
#                 quiet=True,  #  mrcat parameter
#         # Connect inputs
#             'first_scan': ('low_b_dw_scan', _format),  # input mrcat
#             'second_scan': ('high_b_dw_scan', _format),  # input mrcat
#         # Connect outputs
#         pipeline.connect_output('dwi_scan', mrcat, 'out_file')
#         # Check inputs/outputs are connected
#         return pipeline
# 
#     def noddi_fitting_pipeline(self, **name_maps):  # @UnusedVariable
#         """
#         Creates a ROI in which the NODDI processing will be performed
# 
#         Parameters
#         ----------
#         single_slice: Int
#             If provided the processing is only performed on a single slice
#             (for testing)
#         noddi_model: Str
#             Name of the NODDI model to use for the fitting
#         nthreads: Int
#             Number of processes to use
#         """
#         pipeline_name = 'noddi_fitting'
#         inputs = [FilesetSpec('bias_correct', nifti_gz_format),
#                   FilesetSpec('grad_dirs', fsl_bvecs_format),
#                   FilesetSpec('bvalues', fsl_bvals_format)]
#         if self.branch('single_slice'):
#             inputs.append(FilesetSpec('eroded_mask', nifti_gz_format))
#         else:
#             inputs.append(FilesetSpec('brain_mask', nifti_gz_format))
#         pipeline = self.new_pipeline(
#             name=pipeline_name,
#             inputs=inputs,
#             outputs=[FilesetSpec('ficvf', nifti_format),
#                      FilesetSpec('odi', nifti_format),
#                      FilesetSpec('fiso', nifti_format),
#                      FilesetSpec('fibredirs_xvec', nifti_format),
#                      FilesetSpec('fibredirs_yvec', nifti_format),
#                      FilesetSpec('fibredirs_zvec', nifti_format),
#                      FilesetSpec('fmin', nifti_format),
#                      FilesetSpec('kappa', nifti_format),
#                      FilesetSpec('error_code', nifti_format)],
#             desc=(
#                 "Creates a ROI in which the NODDI processing will be "
#                 "performed"),
#             citations=[noddi_cite],
#             name_maps=name_maps)
#         # Create node to unzip the nifti files
#         unzip_bias_correct = pipeline.add(
#             "unzip_bias_correct", MRConvert(),
#             requirements=[mrtrix_req.v('3.0rc3')])
#                 out_ext='nii',  #  unzip_bias_correct parameter
#                 quiet=True,  #  unzip_bias_correct parameter
#         unzip_mask = pipeline.add("unzip_mask", MRConvert(),
#                                   requirements=[mrtrix_req.v('3.0rc3')])
#                 out_ext='nii',  #  unzip_mask parameter
#                 quiet=True,  #  unzip_mask parameter
#         # Create create-roi node
#         create_roi = pipeline.add(
#             'create_roi',
#             CreateROI(),
#             requirements=[noddi_req, matlab_req.v('R2015a')],
#             mem_gb=4)
#             'in_file': (unzip_bias_correct, 'out_file'),  # internal create_roi
#             'brain_mask': (unzip_mask, 'out_file'),  # internal create_roi
#         # Create batch-fitting node
#         batch_fit = pipeline.add(
#             "batch_fit", BatchNODDIFitting(),
#             requirements=[noddi_req, matlab_req.v('R2015a')], wall_time=180,
#             mem_gb=8)
#                 model=self.parameter('noddi_model'),  #  batch_fit parameter
#                 nthreads=self.processor.num_processes,  #  batch_fit parameter
#             'roi_file': (create_roi, 'out_file'),  # internal batch_fit
#         # Create output node
#         save_params = pipeline.add(
#             "save_params",
#             SaveParamsAsNIfTI(),
#             requirements=[noddi_req, matlab_req.v('R2015a')],
#             mem_gb=4)
#                 output_prefix='params',  #  save_params parameter
#             'params_file': (batch_fit, 'out_file'),  # internal save_params
#             'roi_file': (create_roi, 'out_file'),  # internal save_params
#         pipeline.connect(unzip_mask, 'out_file', save_params,
#                          'brain_mask_file')
#         # Connect inputs
#             'in_file': ('bias_correct', _format),  # input unzip_bias_correct
#         if pipeline.branch('single_slice'):
#                 'in_file': ('brain_mask', _format),  # input unzip_mask
#         else:
#                 'in_file': ('eroded_mask', _format),  # input unzip_mask
#             'bvecs_file': ('grad_dirs', _format),  # input batch_fit
#             'bvals_file': ('bvalues', _format),  # input batch_fit
#         # Connect outputs
#         pipeline.connect_output('ficvf', save_params, 'ficvf')
#         pipeline.connect_output('odi', save_params, 'odi')
#         pipeline.connect_output('fiso', save_params, 'fiso')
#         pipeline.connect_output('fibredirs_xvec', save_params,
#                                 'fibredirs_xvec')
#         pipeline.connect_output('fibredirs_yvec', save_params,
#                                 'fibredirs_yvec')
#         pipeline.connect_output('fibredirs_zvec', save_params,
#                                 'fibredirs_zvec')
#         pipeline.connect_output('fmin', save_params, 'fmin')
#         pipeline.connect_output('kappa', save_params, 'kappa')
#         pipeline.connect_output('error_code', save_params, 'error_code')
#         # Check inputs/outputs are connected
#         return pipeline



import os.path
from nipype.interfaces.base import (
    traits, InputMultiPath, File, TraitedSpec, isdefined)
from nipype.interfaces.mrtrix3.reconst import (
    MRTrix3Base, MRTrix3BaseInputSpec)
from nipype.interfaces.mrtrix3.preprocess import (
    ResponseSD as NipypeResponseSD,
    ResponseSDInputSpec as NipypeResponseSDInputSpec)
from arcana.utils import split_extension


class Fod2FixelInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class Fod2FixelOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class Fod2Fixel(MRTrix3Base):
    """Fod2Fixel"""

    _cmd = "fod2fixel"
    input_spec = Fod2FixelInputSpec
    output_spec = Fod2FixelOutputSpec

    def _list_outputs(self):

        outputs = self.output_spec().get()
        return outputs


class Fixel2VoxelInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class Fixel2VoxelOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class Fixel2Voxel(MRTrix3Base):
    """Fixel2Voxel"""

    _cmd = "fixel2voxel"
    input_spec = Fixel2VoxelInputSpec
    output_spec = Fixel2VoxelOutputSpec

    def _list_outputs(self):

        outputs = self.output_spec().get()
        return outputs


class FixelCorrespondenceInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class FixelCorrespondenceOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class FixelCorrespondence(MRTrix3Base):
    """FixelCorrespondence"""

    _cmd = "fixelcorrespondence"
    input_spec = FixelCorrespondenceInputSpec
    output_spec = FixelCorrespondenceOutputSpec

    def _list_outputs(self):

        outputs = self.output_spec().get()
        return outputs


class FixelCFEStatsInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class FixelCFEStatsOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class FixelCFEStats(MRTrix3Base):
    """FixelCFEStats"""

    _cmd = "fixelcfestats"
    input_spec = FixelCFEStatsInputSpec
    output_spec = FixelCFEStatsOutputSpec

    def _list_outputs(self):

        outputs = self.output_spec().get()
        return outputs


class TckSiftInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class TckSiftOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class TckSift(MRTrix3Base):
    """TckSift"""

    _cmd = "tcksift"
    input_spec = TckSiftInputSpec
    output_spec = TckSiftOutputSpec

    def _list_outputs(self):

        outputs = self.output_spec().get()
        return outputs


class Warp2MetricInputSpec(MRTrix3BaseInputSpec):
    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input files')
    out_file = File(genfile=True, argstr='%s', desc=(""), position=-1)


class Warp2MetricOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc=(""))


class Warp2Metric(MRTrix3Base):
    """Warp2Metric"""

    _cmd = "warp2metric"
    input_spec = Warp2MetricInputSpec
    output_spec = Warp2MetricOutputSpec

    def _list_outputs(self):

        outputs = self.output_spec().get()
        return outputs


class AverageReponseInputSpec(MRTrix3BaseInputSpec):

    in_files = InputMultiPath(
        File(exists=True), argstr='%s', mandatory=True,
        position=0, desc="Average response")

    out_file = File(
        genfile=True, argstr='%s', position=-1,
        desc=("the output spherical harmonics coefficients image"))


class AverageReponseOutputSpec(TraitedSpec):

    out_file = File(exists=True, desc='the output response file')


class AverageResponse(MRTrix3Base):

    _cmd = 'average_response'
    input_spec = AverageReponseInputSpec
    output_spec = AverageReponseOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_filename(self, name):
        if name == 'out_file':
            fname = self._gen_outfilename()
        else:
            assert False
        return fname

    def _gen_outfilename(self):
        if isdefined(self.inputs.out_file):
            filename = self.inputs.out_file
        else:
            base, ext = split_extension(
                os.path.basename(self.inputs.in_files[0]))
            filename = os.path.join(os.getcwd(),
                                    "{}_avg{}".format(base, ext))
        return filename


class EstimateFODInputSpec(MRTrix3BaseInputSpec):

    algorithm = traits.Enum('csd', 'msmt_csd', argstr='%s', mandatory=True,
                            position=0,
                            desc="Algorithm used for CSD estimation")

    in_file = File(exists=True, argstr='%s', mandatory=True, position=-3,
                   desc='input diffusion weighted images')
    response = File(
        exists=True, argstr='%s', mandatory=True, position=-2,
        desc=('a text file containing the diffusion-weighted signal response '
              'function coefficients for a single fibre population'))
    out_file = File(
        'fods.mif', argstr='%s', mandatory=True, position=-1,
        usedefault=True, desc=('the output spherical harmonics coefficients'
                               ' image'))

    # DW Shell selection parameters
    shell = traits.List(traits.Float, sep=',', argstr='-shell %s',
                        desc='specify one or more dw gradient shells')

    # Spherical deconvolution parameters
    max_sh = traits.Int(8, argstr='-lmax %d',
                        desc='maximum harmonic degree of response function')
    in_mask = File(exists=True, argstr='-mask %s',
                   desc='provide initial mask image')
    in_dirs = File(
        exists=True, argstr='-directions %s',
        desc=('specify the directions over which to apply the non-negativity '
              'constraint (by default, the built-in 300 direction set is '
              'used). These should be supplied as a text file containing the '
              '[ az el ] pairs for the directions.'))
    sh_filter = File(
        exists=True, argstr='-filter %s',
        desc=('the linear frequency filtering parameters used for the initial '
              'linear spherical deconvolution step (default = [ 1 1 1 0 0 ]). '
              'These should be supplied as a text file containing the '
              'filtering coefficients for each even harmonic order.'))

    neg_lambda = traits.Float(
        1.0, argstr='-neg_lambda %f',
        desc=('the regularisation parameter lambda that controls the strength'
              ' of the non-negativity constraint'))
    thres = traits.Float(
        0.0, argstr='-threshold %f',
        desc=('the threshold below which the amplitude of the FOD is assumed '
              'to be zero, expressed as an absolute amplitude'))

    n_iter = traits.Int(
        50, argstr='-niter %d', desc=('the maximum number of iterations '
                                      'to perform for each voxel'))


class EstimateFODOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='the output response file')


class EstimateFOD(MRTrix3Base):

    """
    Convert diffusion-weighted images to tensor images

    Note that this program makes use of implied symmetries in the diffusion
    profile. First, the fact the signal attenuation profile is real implies
    that it has conjugate symmetry, i.e. Y(l,-m) = Y(l,m)* (where * denotes
    the complex conjugate). Second, the diffusion profile should be
    antipodally symmetric (i.e. S(x) = S(-x)), implying that all odd l
    components should be zero. Therefore, this program only computes the even
    elements.

    Note that the spherical harmonics equations used here differ slightly from
    those conventionally used, in that the (-1)^m factor has been omitted.
    This should be taken into account in all subsequent calculations.
    The spherical harmonic coefficients are stored as follows. First, since
    the signal attenuation profile is real, it has conjugate symmetry, i.e.
    Y(l,-m) = Y(l,m)* (where * denotes the complex conjugate). Second, the
    diffusion profile should be antipodally symmetric (i.e. S(x) = S(-x)),
    implying that all odd l components should be zero. Therefore, only the
    even elements are computed.

    Note that the spherical harmonics equations used here differ slightly from
    those conventionally used, in that the (-1)^m factor has been omitted.
    This should be taken into account in all subsequent calculations.
    Each volume in the output image corresponds to a different spherical
    harmonic component. Each volume will correspond to the following:

    volume 0: l = 0, m = 0
    volume 1: l = 2, m = -2 (imaginary part of m=2 SH)
    volume 2: l = 2, m = -1 (imaginary part of m=1 SH)
    volume 3: l = 2, m = 0
    volume 4: l = 2, m = 1 (real part of m=1 SH)
    volume 5: l = 2, m = 2 (real part of m=2 SH)
    etc...



    Example
    -------

    >>> import nipype.interfaces.mrtrix3 as mrt
    >>> fod = mrt.EstimateFOD()
    >>> fod.inputs.in_file = 'dwi.mif'
    >>> fod.inputs.response = 'response.txt'
    >>> fod.inputs.in_mask = 'mask.nii.gz'
    >>> fod.inputs.grad_fsl = ('bvecs', 'bvals')
    >>> fod.cmdline                               # doctest: +ELLIPSIS +ALLOW_UNICODE
    'dwi2fod -fslgrad bvecs bvals -mask mask.nii.gz dwi.mif response.txt\
 fods.mif'
    >>> fod.run()                                 # doctest: +SKIP
    """

    _cmd = 'dwi2fod'
    input_spec = EstimateFODInputSpec
    output_spec = EstimateFODOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs['out_file'] = os.path.abspath(self.inputs.out_file)
        return outputs




import os.path
import warnings
from string import Template
from nibabel import load
from nipype.interfaces.base import (
    File, traits, TraitedSpec, BaseInterface, BaseInterfaceInputSpec,
    Directory)
from glob import glob
from nipype.interfaces.fsl.base import (FSLCommand, FSLCommandInputSpec)
import logging
from nipype.interfaces.base import isdefined
import nibabel as nib
import numpy as np
import ast
import subprocess as sp
import scipy
from random import shuffle
import shutil


warn = warnings.warn
warnings.filterwarnings('always', category=UserWarning)

feat_template_path = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "resources", 'temp.fsf')
optiBET_path = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            'resources', 'bash', 'optiBET.sh'))

logger = logging.getLogger('arcana')


class MelodicL1FSFInputSpec(BaseInterfaceInputSpec):

    output_dir = traits.String(desc="Output directory", mandatory=True)

    tr = traits.Float(desc="TR", mandatory=True)
    brain_thresh = traits.Float(
        desc="Brain/Background threshold %", mandatory=True)
    dwell_time = traits.Float(desc="EPI dwell time", mandatory=True)
    te = traits.Float(desc="TE", mandatory=True)
    unwarp_dir = traits.Str(desc="Unwrap direction", mandatory=True)
    sfwhm = traits.Float(desc="Smoothing FWHM", mandatory=True)
    fmri = File(exists=True, desc="4D functional data", mandatory=True)
    fmri_ref = File(
        exists=True, desc="reference functional data", mandatory=True)
    fmap = File(exists=True, desc="fieldmap file", mandatory=True)
    fmap_mag = File(
        exists=True, desc="fieldmap magnitude file", mandatory=True)
    structural = File(exist=True, desc="Structural image", mandatory=True)
    high_pass = traits.Float(desc="high pass cutoff (s)", mandatory=True)


class MelodicL1FSFOutputSpec(TraitedSpec):

    fsf_file = File(exists=True)


class MelodicL1FSF(BaseInterface):

    def get_vols(self, nifti):
        img = load(nifti)
        if len(img.shape) < 4:
            return 1
        else:
            return img.shape[3]

    input_spec = MelodicL1FSFInputSpec
    output_spec = MelodicL1FSFOutputSpec

    def _run_interface(self, runtime):
        template_file = open(feat_template_path)
        template = Template(template_file.read())
        template_file.close()
#        inputs = self.input_spec().get()
        d = {}
        d['outputdir'] = self.inputs.output_dir
        d['tr'] = self.inputs.tr
        d['volumes'] = self.get_vols(self.inputs.fmri)
        d['del_volume'] = 0
        d['brain_thresh'] = self.inputs.brain_thresh
        d['epi_dwelltime'] = self.inputs.dwell_time
        d['epi_te'] = self.inputs.te
        d['unwarp_dir'] = self.inputs.unwarp_dir
        d['smoothing_fwhm'] = self.inputs.sfwhm
        d['highpass_cutoff'] = self.inputs.high_pass
        d['fmri_file'] = self.inputs.fmri
        d['fmri_reference'] = self.inputs.fmri_ref
        d['fieldmap'] = self.inputs.fmap
        d['fieldmap_mag'] = self.inputs.fmap_mag
        d['structural'] = self.inputs.structural
        d['fsl_dir'] = os.environ['FSLDIR']
        out = template.substitute(d)

        tt = open('melodic.fsf', 'w')
        tt.write(out)
        tt.close()
        return runtime

    def _list_outputs(self):

        outputs = self.output_spec().get()
        outputs['fsf_file'] = os.path.abspath('./melodic.fsf')
        return outputs


class SignalRegressionInputSpec(BaseInterfaceInputSpec):

    fix_dir = Directory(exists=True, desc="Prepared FIX directory",
                        mandatory=True)
    labelled_components = File(
        exists=True, mandatory=True,
        desc=("Text file with FIX classified components."))
    motion_regression = traits.Bool(desc='Regress 24 motion paremeters.',
                                    default=False)
    highpass = traits.Float(
        desc='Whether or not to high pass the motion parameters', default=None)
    customRegressors = File(exists=True, default=None,
                            desc='File containing custom regressors.')


class SignalRegressionOutputSpec(TraitedSpec):

    output = File(exists=True, desc="cleaned output")


class SignalRegression(BaseInterface):

    input_spec = SignalRegressionInputSpec
    output_spec = SignalRegressionOutputSpec

    def _run_interface(self, runtime):

        im2filt = self.inputs.fix_dir+'/filtered_func_data.nii.gz'
        ref = nib.load(im2filt)
        components = []
        with open(self.inputs.labelled_components, 'r') as f:
            for line in f:
                components.append(line)
        bad_components = ast.literal_eval(components[-1].strip())
        bad_components = [x-1 for x in bad_components]
        im2filt = nib.load(im2filt)
        if self.inputs.highpass:
            hdr = im2filt.header
            sa = hdr.structarr
            TR = sa['pixdim'][4]
            print('Repetition time from the header: {} sec'.format(str(TR)))
        else:
            TR = None
        im2filt = im2filt.get_data()
        [x, y, z, t] = im2filt.shape
        im2filt = np.reshape(im2filt, (x*y*z, t), order='F').T
        ICA = self.normalise(np.loadtxt(self.inputs.fix_dir+'/melodic_mix'))
        if self.inputs.motion_regression:
            mp = self.inputs.fix_dir+'/mc/prefiltered_func_data_mcf.par'
            motion_confounds = self.create_motion_confounds(
                mp, self.inputs.highpass, TR=TR)
            ICA = ICA - (np.dot(motion_confounds,
                                np.dot(np.linalg.pinv(motion_confounds), ICA)))
            im2filt = (im2filt-np.dot(
                motion_confounds, np.dot(np.linalg.pinv(motion_confounds),
                                         im2filt)))
        if self.inputs.customRegressors:
            cr = np.loadtxt(self.inputs.customRegressors)
            if cr.shape[0] != t:
                print (
                    'custom regressors and input image have a different '
                    'time lenght. They will not be used for the regression.')
            else:
                cr = self.normalise(cr)
                im2filt = im2filt - np.dot(cr, np.dot(np.linalg.pinv(cr),
                                                      im2filt))

        betaICA = np.dot(np.linalg.pinv(ICA), im2filt)
        im2filt = im2filt - np.dot(ICA[:, bad_components],
                                   betaICA[bad_components, :])
        im_filt = np.reshape(im2filt.T, (x, y, z, t), order='F')
        im2save = nib.Nifti1Image(im_filt, affine=ref.get_affine())
        nib.save(
            im2save, self.inputs.fix_dir+'/filtered_func_data_clean.nii.gz')

        return runtime

    def create_motion_confounds(self, mp, hp, TR=None):

        confounds = np.loadtxt(mp)
        confounds = self.normalise(
            np.hstack((confounds, np.vstack((np.zeros(6),
                                             np.diff(confounds, axis=0))))))
        confounds = self.normalise(
            np.hstack((confounds, np.square(confounds))))
        if hp == 0:
            confounds = scipy.signal.detrend(confounds, axis=0, type='linear')
        elif hp > 0:
            im2save = nib.Nifti1Image(
                np.reshape(confounds.T, (confounds.shape[1], 1, 1,
                                         confounds.shape[0]), order='F'),
                affine=np.eye(4))
            nib.save(im2save, self.inputs.fix_dir+'/mc/mc_par_conf.nii.gz')
            cmd = (
                'fslmaths {0}/mc_par_conf.nii.gz -bptf {1} -1 '
                '{0}/mc_par_conf_hp'.format(self.inputs.fix_dir+'/mc',
                                            str(0.5*float(hp)/TR)))
            sp.check_output(cmd, shell=True)
            confounds_hp = (nib.load(
                self.inputs.fix_dir+'/mc/mc_par_conf_hp.nii.gz')).get_data()
            confounds = self.normalise(
                np.reshape(confounds_hp, (confounds.shape[1],
                                          confounds.shape[0]), order='F').T)

        return confounds

    def normalise(self, params):

        params = (params - np.mean(params, axis=0))/np.std(params, axis=0,
                                                           ddof=1)
        return params

    def _list_outputs(self):

        outputs = self._outputs().get()
        outputs['output'] = (
            self.inputs.fix_dir+'/filtered_func_data_clean.nii.gz')
        return outputs


class FSLFIXInputSpec(FSLCommandInputSpec):
    _xor_parameters = ('classification', 'regression', 'all')
    _xor_input_files = ('feat_dir', 'labelled_component')
    feat_dir = Directory(
        exists=True, argstr="%s", position=1, xor=_xor_input_files,
        desc="Input melodic preprocessed directory")
    train_data = File(exists=True, argstr="%s", position=2,
                      desc="Training file")
    regression = traits.Bool(desc='Regress previously classified components.',
                             position=0, argstr="-a", xor=_xor_parameters)
    classification = traits.Bool(
        desc='Components classification without regression.', position=0,
        argstr="-c", xor=_xor_parameters)
    all = traits.Bool(
        desc='Components classification and regression.', position=0,
        argstr="-f", xor=_xor_parameters)
    component_threshold = traits.Int(
        argstr="%d", mandatory=True, position=3,
        desc="threshold for the number of components")
    labelled_component = File(
        exists=True, argstr="%s", position=1, xor=_xor_input_files,
        desc=("Text file with classified components. This file is mandatory if"
              "you choose regression only."))
    motion_reg = traits.Bool(argstr='-m', desc="motion parameters regression")
    highpass = traits.Float(
        argstr='-h %f', desc='apply highpass of the motion '
        'confound with <highpass> being full-width (2*sigma) in seconds.')


class FSLFIXOutputSpec(TraitedSpec):
    output = File(exists=True, desc="cleaned output")
    label_file = File(exists=True, desc="labelled components")


class FSLFIX(FSLCommand):

    _cmd = 'fix'
    input_spec = FSLFIXInputSpec
    output_spec = FSLFIXOutputSpec
    text_ext = '.txt'

    def _list_outputs(self):
        outputs = self.output_spec().get()
        # print(self.inputs.feat_dir+'./filtered_func_data_clean.nii*')
        if self.inputs.all:
            outputs['output'] = self._gen_filename('out_file')
            outputs['label_file'] = self._gen_filename('label_file')
        elif self.inputs.classification:
            outputs['label_file'] = self._gen_filename('label_file')
        elif self.inputs.regression:
            outputs['output'] = self._gen_filename('out_file')
        else:
            outputs['output'] = self._gen_filename('out_file')
            outputs['label_file'] = self._gen_filename('label_file')
        return outputs

    def _gen_filename(self, name):
        if isdefined(self.inputs.feat_dir):
            cwd = self.inputs.feat_dir
        elif isdefined(self.inputs.labelled_component):
            cwd = '/'.join(self.inputs.labelled_component.split('/')[:-1])

        if name == 'out_file':
            fname = cwd+glob('/filtered_func_data_clean.nii*')[0]
        elif name == 'label_file':
            fid = os.path.basename(self.inputs.train_data).split('.RData')[0]
            thr = str(self.inputs.component_threshold)
            fname = cwd+'/fix4melview_'+fid+'_thr'+thr+self.text_ext
        else:
            assert False
        return fname


class FSLFixTrainingInputSpec(FSLCommandInputSpec):
    training = traits.Bool(mandatory=True, argstr="-t", position=1)
    list_dir = traits.List(mandatory=True, argstr="%s", position=-1,
                           desc="Input feat preprocessed directory")
    outname = traits.Str(mandatory=True, argstr="%s", position=2,
                         desc="output name")


class FSLFixTrainingOutputSpec(TraitedSpec):
    training_set = File(exists=True, desc="training set")


class FSLFixTraining(FSLCommand):

    _cmd = 'fix'
    input_spec = FSLFixTrainingInputSpec
    output_spec = FSLFixTrainingOutputSpec
    ext = '.RData'

    def _list_outputs(self):
        outputs = self.output_spec().get()
        # print self.inputs.feat_dir+'./filtered_func_data_clean.nii*'
        outputs['training_set'] = os.path.join(
            os.getcwd(), self._gen_filename('train_file'))
        return outputs

    def _gen_filename(self, name):
        if name == 'train_file':
            fid = os.path.basename(self.inputs.outname)
            fname = fid + self.ext
        else:
            assert False
        return fname


class CheckLabelFileInputSpec(BaseInterfaceInputSpec):
    in_list = traits.List(desc='melodic directory', mandatory=True)


class CheckLabelFileOutputSpec(TraitedSpec):
    out_list = traits.List(desc="List of melodic dirs that contain "
                                "label file")


class CheckLabelFile(BaseInterface):
    input_spec = CheckLabelFileInputSpec
    output_spec = CheckLabelFileOutputSpec

    def _run_interface(self, runtime):
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        out = []

        for s in self.inputs.in_list:
            if 'hand_labels_noise.txt' in os.listdir(s):
                with open(s+'/hand_labels_noise.txt', 'r') as f:
                    for line in f:
                        el = [x for x in
                              line.strip().strip('[').strip(']').split(',')]
                    f.close()
                    ic = sorted(glob(s+'/report/f*.txt'))
                    if [x for x in el if int(x) > len(ic)]:
                        logger.warning('Subject {} has wrong number of '
                                       'components in the '
                                       'hand_labels_noise.txt file. It will '
                                       'not be used for the FIX training.'
                                       .format(s))
                    else:
                        out.append(s)

        outputs["out_list"] = out
        return outputs


class FSLSlicesInputSpec(FSLCommandInputSpec):
    im1 = File(mandatory=True, position=0, argstr="%s", desc="First image")
    im2 = File(mandatory=True, position=1, argstr="%s", desc="Second image")
    outname = traits.Str(mandatory=True, argstr="-o %s.gif", position=-1,
                         desc="output name")


class FSLSlicesOutputSpec(TraitedSpec):
    report = File(exists=True, desc=".gif file with im2 overlaid to im1")


class FSLSlices(FSLCommand):

    _cmd = 'slices '
    input_spec = FSLSlicesInputSpec
    output_spec = FSLSlicesOutputSpec
    ext = '.gif'

    def _list_outputs(self):
        outputs = self.output_spec().get()
        # print self.inputs.feat_dir+'./filtered_func_data_clean.nii*'
        outputs['report'] = os.path.join(
            os.getcwd(), self._gen_filename('report'))
        return outputs

    def _gen_filename(self, name):
        if name == 'report':
            fid = os.path.basename(self.inputs.outname)
            fname = fid + self.ext
        else:
            assert False
        return fname


class PrepareFIXTrainingInputSpec(BaseInterfaceInputSpec):

    inputs_list = traits.List()
    epi_number = traits.Int()


class PrepareFIXTrainingOutputSpec(TraitedSpec):

    prepared_dirs = traits.List()


class PrepareFIXTraining(BaseInterface):

    input_spec = PrepareFIXTrainingInputSpec
    output_spec = PrepareFIXTrainingOutputSpec

    def _run_interface(self, runtime):
        
        epi_number = self.inputs.epi_number
        inputs = self.inputs.inputs_list
        fix_dirs = [inputs[x:x+epi_number] for x in
                    np.arange(0, len(inputs)/2, epi_number, dtype=int)]
        labels = [inputs[x:x+epi_number] for x in
                    np.arange(len(inputs)/2, len(inputs), epi_number, dtype=int)]
        self.out_dirs = []
        

        if len(labels) != len(fix_dirs):
            raise Exception('The number of subjects provided is different from'
                            'the number of hand_label_noise files. Fix '
                            'training cannot be performed. Please check.')
        for i, label in enumerate(labels):
            for j in range(epi_number):
                with open(label[j], 'r') as f:
                    if 'not_provided' not in f.readline():
                        shutil.copy2(label[j], fix_dirs[i][j]+'/hand_labels_noise.txt')
                        self.out_dirs.append(fix_dirs[i][j])
        if not self.out_dirs:
            raise Exception('No non-empty hand_labels_noise.txt file found in the fix_dir '
                            'provided. In order to run FIX training at least 25 '
                            'hand_labels_noise.txt files have to be provided. Please '
                            'go through 25 single-subject MELODIC ICA results, create '
                            'those text files and upload them on XNAT. Have a look at '
                            'the documentation to have more information.')

        return runtime

    def _list_outputs(self):
        outputs = self.output_spec().get()
        # print self.inputs.feat_dir+'./filtered_func_data_clean.nii*'
        outputs['prepared_dirs'] = self.out_dirs
        return outputs



from nipype.interfaces.base import (BaseInterface, BaseInterfaceInputSpec,
                                    traits, File, TraitedSpec, Directory)


class GenFSBrainMasksInputSpec(BaseInterfaceInputSpec):

    fs_directory = Directory(exists=True, mandatory=True, desc='Directory with'
                             'freesurfer results.')


class GenFSBrainMasksOutputSpec(TraitedSpec):

    totGM = File(exists=True, desc='Total grey matter mask')
    corticalGM = File(exists=True, desc='Cortical grey matter mask')
    totWM = File(exists=True, desc='Total white matter mask')
    subcorticalGM = File(exists=True, desc='Subcortical grey matter mask')
    totCSF = File(exists=True, desc='Total CSF mask')
    totBrainVol = File(exists=True, desc='Total brain volume mask')
    totBrainVolNoVent = File(exists=True, desc='Total brain volume mask'
                             'without ventricles')
    Brainstem = File(exists=True, desc='Brainstem mask')


class GenFSBrainMasks(BaseInterface):

    input_spec = GenFSBrainMasksInputSpec
    output_spec = GenFSBrainMasksOutputSpec

    def _run_interface(self, runtime):
        pass
